/* Domination checking routines */

# include <stdio.h>
# include <stdlib.h>
# include <math.h>

# include "global.h"
# include "rand.h"

/* Routine for non-domination checking (Pareto and equitable efficiency).
   It will return the following values
   1 if a dominates b
   -1 if b dominates a
   0 if both a and b are non-dominated */

int check_dominance (individual *a, individual *b)
{
    if (a->constr_violation<0 && b->constr_violation<0)
    {
        if (a->constr_violation > b->constr_violation)
        {
            return (1);
        }
        else
        {
            if (a->constr_violation < b->constr_violation)
            {
                return (-1);
            }
            else
            {
                return (0);
            }
        }
    }
    else
    {
        if (a->constr_violation < 0 && b->constr_violation == 0)
        {
            return (-1);
        }
        else
        {
            if (a->constr_violation == 0 && b->constr_violation <0)
            {
                return (1);
            }
            else
            {
                if (efficiency == 0)
                {
                	return check_pareto_dominance(a, b);
                }
                else
                {
                	return check_equitable_dominance(a, b);
                }
            }
        }
    }
}




/* Routine that actually checks Pareto dominance.
   It will return the following values
   1 if a dominates b
   -1 if b dominates a
   0 if both a and b are non-dominated */
int check_pareto_dominance(individual *a, individual *b)
{
	int i;
	int flag1 = 0;
	int flag2 = 0;
    for (i=0; i<nobj; ++i)
	{
		if (a->obj[i] < b->obj[i])
		{
			flag1 = 1;

		}
		else
		{
			if (a->obj[i] > b->obj[i])
			{
				flag2 = 1;
			}
		}
	}
	if (flag1==1 && flag2==0)
	{
		return (1);
	}
	else
	{
		if (flag1==0 && flag2==1)
		{
			return (-1);
		}
		else
		{
			return (0);
		}
	}
}




/* Routine that actually checks equitable dominance.
   It will return the following values
   1 if a dominates b
   -1 if b dominates a
   0 if both a and b are non-dominated */
int check_equitable_dominance(individual *a, individual *b)
{
	int i, j;
	int flag1 = 0;
	int flag2 = 0;
	double temp;
	double *obj_arr_a = (double *) malloc(nobj*sizeof(double));
	double *obj_arr_b = (double *) malloc(nobj*sizeof(double));

	for(i=0; i<nobj; ++i)
	{
		obj_arr_a[i] = a->obj[i];
		obj_arr_b[i] = b->obj[i];
	}

	/* build ascending ordering maps */
	quicksort_arr(obj_arr_a, 0, nobj-1);
	quicksort_arr(obj_arr_b, 0, nobj-1);

	/* reverse ordering maps */
	for (i=0, j=nobj-1; i<j; ++i, --j)
	{
		temp = obj_arr_a[j];
		obj_arr_a[j] = obj_arr_a[i];
		obj_arr_a[i] = temp;
		temp = obj_arr_b[j];
		obj_arr_b[j] = obj_arr_b[i];
		obj_arr_b[i] = temp;
	}

	/* build cumulative ordering maps */
	for (i=1; i<nobj; ++i)
	{
		obj_arr_a[i] += obj_arr_a[i-1];
		obj_arr_b[i] += obj_arr_b[i-1];
	}

	/* check domination */
	for (i=0; i<nobj; ++i)
	{
		if (obj_arr_a[i] < obj_arr_b[i])
		{
			flag1 = 1;

		}
		else
		{
			if (obj_arr_a[i] > obj_arr_b[i])
			{
				flag2 = 1;
			}
		}
	}


	free(obj_arr_a);
	free(obj_arr_b);

	if (flag1==1 && flag2==0)
	{
		return (1);
	}
	else
	{
		if (flag1==0 && flag2==1)
		{
			return (-1);
		}
		else
		{
			return (0);
		}
	}

}



