/* Nond-domination based selection routines */

# include <stdio.h>
# include <stdlib.h>
# include <math.h>
# include <assert.h>
# include <stdbool.h>

# include "global.h"
# include "rand.h"


/* Routine to perform non-dominated sorting */
void fill_nondominated_sort (population *mixed_pop, population *new_pop, population *f0_pop)
{
    int flag;
    int i, j, k;
    int nbfree, nbproper;
    int end;
    int front_size = 0;
    int archieve_size = 0;
    double rank = 1.0;
    list *pool;
    list *elite;
    list *elite0;
    list *elite1;
    list *temp1, *temp2;
    pool = (list *)malloc(sizeof(list));
    elite = (list *)malloc(sizeof(list));
    elite0 = (list *)malloc(sizeof(list));
    elite1 = (list *)malloc(sizeof(list));
    pool->index = -1;
    pool->parent = NULL;
    pool->child = NULL;
    elite->index = -1;
    elite->parent = NULL;
    elite->child = NULL;
    elite0->index = -1;
    elite0->parent = NULL;
    elite0->child = NULL;
    elite1->index = -1;
	elite1->parent = NULL;
	elite1->child = NULL;
    temp1 = pool;
    for (i=0; i<2*popsize; i++)
    {
        insert (temp1,i);
        temp1 = temp1->child;
    }
    i=0;
    do
    {
        temp1 = pool->child;
        insert (elite, temp1->index);
        front_size = 1;
        temp2 = elite->child;
        temp1 = del (temp1);
        temp1 = temp1->child;
        do
        {
            temp2 = elite->child;
            if (temp1==NULL)
            {
                break;
            }
            do
            {
                end = 0;
                flag = check_dominance (&(mixed_pop->ind[temp1->index]), &(mixed_pop->ind[temp2->index]));
                if (flag == 1)
                {
                    insert (pool, temp2->index);
                    temp2 = del (temp2);
                    front_size--;
                    temp2 = temp2->child;
                }
                if (flag == 0)
                {
                    temp2 = temp2->child;
                }
                if (flag == -1)
                {
                    end = 1;
                }
            }
            while (end!=1 && temp2!=NULL);
            if (flag == 0 || flag == 1)
            {
                insert (elite, temp1->index);
                front_size++;
                temp1 = del (temp1);
            }
            temp1 = temp1->child;
        }
        while (temp1 != NULL);
        temp2 = elite->child;
        j=i;
        if ( (archieve_size+front_size) <= popsize)
        {
        	if (app_restr_all_fronts || rank<1.5)
            {
        		nbproper = proper_fill(mixed_pop, f0_pop, front_size, elite, elite0, elite1);
				for (k=0; k<nbproper; ++k)
				{
					copy_ind (&f0_pop->ind[k], &new_pop->ind[i]);
					new_pop->ind[i].rank = rank - 0.5;
					archieve_size+=1;
					i+=1;
				}
				if (nbproper > 0)
				{
					assign_crowding_distance_indices (new_pop, j, i-1);
				}
				j = i;
				temp2 = elite1 -> child;
				while (temp2 != NULL)
				{
					copy_ind (&mixed_pop->ind[temp2->index], &new_pop->ind[i]);
					new_pop->ind[i].rank = rank;
					archieve_size+=1;
					temp2 = temp2->child;
					i+=1;
				}
				if (i-j > 0)
				{
					assign_crowding_distance_indices (new_pop, j, i-1);
				}
				rank += 1.0;
            }
            else /*app_restr_all_fronts==0 && rank>1*/
            {
            	while (temp2 != NULL)
				{
					copy_ind (&mixed_pop->ind[temp2->index], &new_pop->ind[i]);
					new_pop->ind[i].rank = rank;
					archieve_size+=1;
					temp2 = temp2->child;
					i+=1;
				}
            	if (i-j > 0)
				{
					assign_crowding_distance_indices (new_pop, j, i-1);
				}
				rank += 1.0;
            }
        }
        else
        {
        	if (app_restr_all_fronts || rank<1.5)
        	{
        		nbfree = popsize - archieve_size; /* number of individuals that fit into new population */
        		nbproper = proper_fill(mixed_pop, f0_pop, front_size, elite, elite0, elite1); /* return size of front 0 */
        		/* elite0 contains all front (rank-0.5) individuals */
        		/* elite1 contains all front (rank) individuals that are not part of front (rank-0.5) */
				if (nbproper<=nbfree)
				{ /* front 0 fits completely into new population */
					for (k=0; k<nbproper; ++k)
					{
						copy_ind (&f0_pop->ind[k], &new_pop->ind[i]);
						new_pop->ind[i].rank = rank - 0.5;
						archieve_size+=1;
						i+=1;
					}
					if (nbproper > 0)
					{
						assign_crowding_distance_indices (new_pop, j, i-1);
					}
					j = i;
					nbfree = popsize - archieve_size;
					if (nbfree > 0)
					{
						crowding_fill (mixed_pop, new_pop, i, front_size-nbproper, elite1); /* take rest from front 1 */
					}
					for (; j<popsize; j++)
					{
						new_pop->ind[j].rank = rank;
					}
				}
				else
				{
					crowding_fill (f0_pop, new_pop, i, nbproper, elite0); /* take individuals from front 0 with highest crowding distance */
					for (j=i; j<popsize; ++j)
					{
						new_pop->ind[j].rank = rank - 0.5;
					}
				}
        	}
        	else /*app_restr_all_front==0 && rank>1*/
        	{
				crowding_fill (mixed_pop, new_pop, i, front_size, elite);/*..die besten auswÃ¤hlen...*/
				for (j=i; j<popsize; ++j)
				{
					new_pop->ind[j].rank = rank;
				}
        	}
			archieve_size = popsize;
        }
        temp2 = elite->child;
        do
        {
            temp2 = del (temp2);
            temp2 = temp2->child;
        }
        while (elite->child !=NULL);
        temp2 = elite0->child;
        while (temp2 != NULL)
		{
			temp2 = del (temp2);
			temp2 = temp2->child;
		}
		temp2 = elite1->child;
		while (temp2 != NULL)
		{
			temp2 = del (temp2);
			temp2 = temp2->child;
		}
    }
    while (archieve_size < popsize);
    while (pool!=NULL)
    {
        temp1 = pool;
        pool = pool->child;
        free (temp1);
    }
    while (elite!=NULL)
    {
        temp1 = elite;
        elite = elite->child;
        free (temp1);
    }
    while (elite0!=NULL)
	{
		temp1 = elite0;
		elite0 = elite0->child;
		free (temp1);
	}
    return;
    while (elite1!=NULL)
	{
		temp1 = elite1;
		elite1 = elite1->child;
		free (temp1);
	}
	return;
}

/* Routine to fill a population with individuals in the decreasing order of crowding distance */
void crowding_fill (population *mixed_pop, population *new_pop, int count, int front_size, list *elite)
{
    int *dist;
    list *temp;
    int i, j;
    assign_crowding_distance_list (mixed_pop, elite->child, front_size);
    dist = (int *)malloc(front_size*sizeof(int));
    temp = elite->child;
    for (j=0; j<front_size; j++)
    {
        dist[j] = temp->index;
        temp = temp->child;
    }
    quicksort_dist (mixed_pop, dist, front_size);
    for (i=count, j=front_size-1; i<popsize; i++, j--)
    {
        copy_ind(&mixed_pop->ind[dist[j]], &new_pop->ind[i]);
    }
    free (dist);
    return;
}

/* Routine to find individuals that are optimal with respect to Geoffrion/tradeoff bound. */
int proper_fill (population *mixed_pop, population *f0_pop, int front_size, list *elite, list *elite0, list *elite1)
{
	list *temp = elite;
	list *temp0 = elite0;
	list *temp1 = elite1; /* elite1 contains all front 1 individuals that are not part of front 0 */
	int test_ind;
	int i;
	int nbprop = 0; /* number of proper individuals */
	int opt = 0;
	for (i=0; i<front_size; ++i)
	{
		temp = temp->child;
		test_ind = temp->index; /* individual to be checked */
		temp = del(temp);
		switch (restr)
		{
			case 0: opt = false; /* no test required (standard nsga ii algorithm)*/
					break;
			case 1: opt = test_geoffrion(mixed_pop, front_size, elite, test_ind); /* if true, solution is Geoffrion M proper */
					break;
			default: exit (1);
		}
		if (opt)
		{
			copy_ind(&mixed_pop->ind[test_ind], &f0_pop->ind[nbprop]); /* copy proper individual to f0_pop */
			insert(temp0, nbprop);
			temp0 = temp0->child;
			insert(temp, test_ind);
			temp = temp->child;
			++nbprop;
		}
		else
		{
			insert(temp, test_ind);
			temp = temp->child;
			insert(temp1, test_ind);
			temp1 = temp1->child;
		}
	}
	return nbprop;
}

/* Routine that actually checks Geoffrion M property*/
/* return 1 if candidate is Geoffrion M properly Pareto optimal, 0 otherwise */
int test_geoffrion(population *mixed_pop, int front_size, list *elite, int test_ind)
{
	int i, j, k;
	int flag = 0;
	double a, b;
	list *temp = elite;
	for (i=0; i<front_size-1; ++i) /* compare candidate with every front 1 solution */
	{
		temp = temp->child;
		for (j=0; j<nobj; ++j) /* compare every objective */
		{
			if ( (a=(mixed_pop->ind[test_ind].obj[j] - mixed_pop->ind[temp->index].obj[j])) > 0) /* if current individual is dominated in a certain objective */
			{
				flag = 1;
				for (k=0; k<nobj && flag==1; ++k)
				{
					if ( (b=(mixed_pop->ind[temp->index].obj[k] - mixed_pop->ind[test_ind].obj[k])) > 0)
					{
						if ( a/b <= m_bound)
						{
							flag = 0;
						}
					}
				}
			}
			if (flag == 1)
			{
				return 0;
			}
		}
	}
	return 1;
}

